from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.control.monad import Monad, MonadLambda
from categories.type import Lambda, _, hkt, typeclass

__all__ = (
    'MonadFix',
    'MonadFixLambda',
)


a = TypeVar('a')

m = TypeVar('m')

r = TypeVar('r')


@dataclass(frozen=True)
class MonadFix(Monad[m], typeclass[m]):
    def fix(self, f : Lambda[a, hkt[m, a]], /) -> hkt[m, a]: ...


@dataclass(frozen=True)
class MonadFixLambda(MonadLambda[r], MonadFix[Lambda[r, _]]):
    def fix(self, f : Lambda[a, Lambda[r, a]], /) -> Lambda[r, a]:
        return lambda r, /: (x := lambda: f(x)(r))()



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.type import typeclass

__all__ = (
    'Read',
)


a = TypeVar('a')


@dataclass(frozen=True)
class Read(typeclass[a]):
    def read(self, s : str, /) -> list[tuple[a, str]]: ...



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.control.alternative import Alternative
from categories.control.applicative import Applicative
from categories.control.liftio import LiftIO
from categories.control.monad import Monad
from categories.control.monadplus import MonadPlus
from categories.control.monadtrans import MonadTrans
from categories.data.functor import Functor
from categories.type import IO, Lambda, _, forall, hkt

__all__ = (
    'ReaderT',
    'FunctorReaderT',
    'ApplicativeReaderT',
    'AlternativeReaderT',
    'MonadReaderT',
    'MonadPlusReaderT',
    'MonadTransReaderT',
    'LiftIOReaderT',
)


a = TypeVar('a')

b = TypeVar('b')

m = TypeVar('m')

r = TypeVar('r')


@dataclass(frozen=True)
class ReaderT(forall[r, m, a]):
    run : Lambda[r, hkt[m, a]]


@dataclass(frozen=True)
class FunctorReaderT(Functor[ReaderT[r, m, _]]):
    inst : Functor[m]

    def map(self, f : Lambda[a, b], x : ReaderT[r, m, a], /) -> ReaderT[r, m, b]:
        match self:
            case FunctorReaderT(inst):
                return ReaderT(lambda r, /: inst.map(f, x.run(r)))

    def const(self, x : a, _ : ReaderT[r, m, b], /) -> ReaderT[r, m, a]:
        match self:
            case FunctorReaderT(inst):
                return ReaderT(lambda r, /: inst.const(x, _.run(r)))


@dataclass(frozen=True)
class ApplicativeReaderT(FunctorReaderT[r, m], Applicative[ReaderT[r, m, _]]):
    inst : Applicative[m]

    def pure(self, x : a, /) -> ReaderT[r, m, a]:
        match self:
            case ApplicativeReaderT(inst):
                return ReaderT(lambda _, /: inst.pure(x))

    def apply(self, f : ReaderT[r, m, Lambda[a, b]], x : ReaderT[r, m, a], /) -> ReaderT[r, m, b]:
        match self:
            case ApplicativeReaderT(inst):
                return ReaderT(lambda r, /: inst.apply(f.run(r), x.run(r)))

    def seq(self, _ : ReaderT[r, m, a], x : ReaderT[r, m, b], /) -> ReaderT[r, m, b]:
        match self:
            case ApplicativeReaderT(inst):
                return ReaderT(lambda r, /: inst.seq(_.run(r), x.run(r)))


@dataclass(frozen=True)
class AlternativeReaderT(ApplicativeReaderT[r, m], Alternative[ReaderT[r, m, _]]):
    inst : Alternative[m]

    def empty(self, /) -> ReaderT[r, m, a]:
        match self:
            case AlternativeReaderT(inst):
                return ReaderT(lambda _, /: inst.empty())

    def alt(self, x : ReaderT[r, m, a], y : ReaderT[r, m, a], /) -> ReaderT[r, m, a]:
        match self:
            case AlternativeReaderT(inst):
                return ReaderT(lambda r, /: inst.alt(x.run(r), y.run(r)))


@dataclass(frozen=True)
class MonadReaderT(ApplicativeReaderT[r, m], Monad[ReaderT[r, m, _]]):
    inst : Monad[m]

    def bind(self, m : ReaderT[r, m, a], k : Lambda[a, ReaderT[r, m, b]], /) -> ReaderT[r, m, b]:
        match self:
            case MonadReaderT(inst):
                return ReaderT(lambda r, /: inst.bind(m.run(r), lambda x, /: k(x).run(r)))


@dataclass(frozen=True)
class MonadPlusReaderT(AlternativeReaderT[r, m], MonadReaderT[r, m], MonadPlus[ReaderT[r, m, _]]):
    inst : MonadPlus[m]


@dataclass(frozen=True)
class MonadTransReaderT(MonadTrans[ReaderT[r, _, _]]):
    def lift(self, inst : Monad[m], m : hkt[m, a], /) -> ReaderT[r, m, a]:
        return ReaderT(lambda _, /: m)


@dataclass(frozen=True)
class LiftIOReaderT(MonadReaderT[r, m], LiftIO[ReaderT[r, m, _]]):
    inst : LiftIO[m]

    def liftIO(self, m : IO[a], /) -> ReaderT[r, m, a]:
        match self:
            case LiftIOReaderT(inst):
                return ReaderT(lambda _, /: inst.liftIO(m))



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.type import Lambda, hkt, typeclass

__all__ = (
    'Category',
    'CategoryLambda',
    'arrow',
)


a = TypeVar('a')

b = TypeVar('b')

c = TypeVar('c')

cat = TypeVar('cat')


@dataclass(frozen=True)
class Category(typeclass[cat]):
    def id(self, /) -> hkt[cat, a, a]: ...

    def o(self, f : hkt[cat, b, c], g : hkt[cat, a, b], /) -> hkt[cat, a, c]: ...


@dataclass(frozen=True)
class CategoryLambda(Category[Lambda]):
    def id(self, /) -> Lambda[a, a]:
        return lambda x, /: x

    def o(self, f : Lambda[b, c], g : Lambda[a, b], /) -> Lambda[a, c]:
        return lambda x, /: f(g(x))


def arrow(inst : Category[cat], f : hkt[cat, a, b], g : hkt[cat, b, c], /) -> hkt[cat, a, c]:
    return inst.o(g, f)



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.control.alternative import Alternative, AlternativeIO, AlternativeList
from categories.control.monad import Monad, MonadIO, MonadList
from categories.type import IO, hkt, typeclass

__all__ = (
    'MonadPlus',
    'MonadPlusIO',
    'MonadPlusList',
)


a = TypeVar('a')

m = TypeVar('m')


@dataclass(frozen=True)
class MonadPlus(Alternative[m], Monad[m], typeclass[m]):
    def zero(self, /) -> hkt[m, a]:
        return self.empty()

    def plus(self, x : hkt[m, a], y : hkt[m, a], /) -> hkt[m, a]:
        return self.alt(x, y)


@dataclass(frozen=True)
class MonadPlusIO(AlternativeIO, MonadIO, MonadPlus[IO]):
    async def zero(self, /) -> a:
        raise Exception

    async def plus(self, m : IO[a], m_ : IO[a], /) -> a:
        try:
            return await m
        except BaseException:
            return await m_


@dataclass(frozen=True)
class MonadPlusList(AlternativeList, MonadList, MonadPlus[list]): ...



from __future__ import annotations

from dataclasses import dataclass
from functools import reduce
from typing import TypeVar

from categories.data.semigroup import Semigroup
from categories.type import typeclass

__all__ = (
    'Monoid',
)


a = TypeVar('a')


@dataclass(frozen=True)
class Monoid(Semigroup[a], typeclass[a]):
    def empty(self, /) -> a:
        return self.concat([])

    def concat(self, xs : list[a], /) -> a:
        return reduce(self.append, xs, self.empty())



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.control.monad import Monad, MonadIO
from categories.type import IO, hkt, typeclass

__all__ = (
    'LiftIO',
    'BaseIO',
)


a = TypeVar('a')

m = TypeVar('m')


@dataclass(frozen=True)
class LiftIO(Monad[m], typeclass[m]):
    def liftIO(self, m : IO[a], /) -> hkt[m, a]: ...


@dataclass(frozen=True)
class BaseIO(MonadIO, LiftIO[IO]):
    async def liftIO(self, m : IO[a], /) -> a:
        return await m



from __future__ import annotations

from typing import TypeVar

from categories.type import Expr, Lambda

__all__ = (
    'fix',
    'o',
    'id',
    'const',
    'apply',
    'reverse',
    'flip',
    'on',
    'when',
)


a = TypeVar('a')

b = TypeVar('b')

c = TypeVar('c')


def fix(f : Expr[..., a], /) -> Expr[..., a]:
    return (x := lambda *_: f(x, *_))


def o(f : Lambda[b, c], g : Lambda[a, b], /) -> Lambda[a, c]:
    return lambda x, /: f(g(x))


def id(x : a, /) -> a:
    return x


def const(x : a, /) -> Lambda[b, a]:
    return lambda _, /: x


def apply(f : Lambda[a, b], x : a, /) -> b:
    return f(x)


def reverse(x : a, f : Lambda[a, b], /) -> b:
    return f(x)


def flip(f : Expr[[a, b], c], /) -> Expr[[b, a], c]:
    return lambda x, y, /: f(y, x)


def on(f : Expr[[b, b], c], g : Lambda[a, b], /) -> Expr[[a, a], c]:
    return lambda x, y, /: f(g(x), g(y))


def when(p : bool, f : Lambda[a, a], x : a, /) -> a:
    return f(x) if p else x


from categories.control.alternative import Alternative
from categories.control.applicative import Applicative
from categories.control.monad import Monad
from categories.control.monadplus import MonadPlus
from categories.core import delay, force, let, rec, seq, undefined
from categories.data.fold import Fold
from categories.data.function import fix, o, id, const, apply
from categories.data.functor import Functor
from categories.data.list import cons, head, tail, filter, foldl, foldr, map, scanl, scanr, unfoldr
from categories.data.monoid import Monoid
from categories.data.semigroup import Semigroup
from categories.data.traverse import Traverse
from categories.data.tuple import fst, snd, curry, uncurry, swap
from categories.type import Expr, IO, Lambda, Void

__all__ = (
    'Alternative',
    'Applicative',
    'Monad',
    'MonadPlus',
    'delay',
    'force',
    'let',
    'rec',
    'seq',
    'undefined',
    'Fold',
    'fix',
    'o',
    'id',
    'const',
    'apply',
    'Functor',
    'cons',
    'head',
    'tail',
    'filter',
    'foldl',
    'foldr',
    'map',
    'scanl',
    'scanr',
    'unfoldr',
    'Monoid',
    'Semigroup',
    'Traverse',
    'fst',
    'snd',
    'curry',
    'uncurry',
    'swap',
    'Expr',
    'IO',
    'Lambda',
    'Void'
)



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.control.alternative import Alternative
from categories.control.applicative import Applicative
from categories.control.monad import Monad
from categories.control.monadfail import MonadFail
from categories.control.monadplus import MonadPlus
from categories.data.functor import Functor
from categories.type import Lambda, forall

__all__ = (
    'Maybe',
    'Nothing',
    'Just',
    'FunctorMaybe',
    'ApplicativeMaybe',
    'AlternativeMaybe',
    'MonadMaybe',
    'MonadFailMaybe',
    'MonadPlusMaybe',
)


a = TypeVar('a')

b = TypeVar('b')


@dataclass(frozen=True)
class Nothing: ...


@dataclass(frozen=True)
class Just(forall[a]):
    x : a


Maybe = Nothing | Just[a]


@dataclass(frozen=True)
class FunctorMaybe(Functor[Maybe]):
    def map(self, f : Lambda[a, b], m : Maybe[a], /) -> Maybe[b]:
        match m:
            case Nothing():
                return Nothing()
            case Just(x):
                return Just(f(x))


@dataclass(frozen=True)
class ApplicativeMaybe(FunctorMaybe, Applicative[Maybe]):
    def pure(self, x : a, /) -> Maybe[a]:
        return Just(x)

    def apply(self, m : Maybe[Lambda[a, b]], m_ : Maybe[a], /) -> Maybe[b]:
        match m:
            case Nothing():
                return Nothing()
            case Just(f):
                return self.map(f, m_)


@dataclass(frozen=True)
class AlternativeMaybe(ApplicativeMaybe, Alternative[Maybe]):
    def empty(self, /) -> Maybe[a]:
        return Nothing()

    def alt(self, m : Maybe[a], m_ : Maybe[a], /) -> Maybe[a]:
        match m:
            case Nothing():
                return m_
            case Just(_):
                return m


@dataclass(frozen=True)
class MonadMaybe(ApplicativeMaybe, Monad[Maybe]):
    def bind(self, m : Maybe[a], k : Lambda[a, Maybe[b]], /) -> Maybe[b]:
        match m:
            case Nothing():
                return Nothing()
            case Just(x):
                return k(x)


@dataclass(frozen=True)
class MonadFailMaybe(MonadMaybe, MonadFail[Maybe]):
    def fail(self, _ : str, /) -> Maybe[a]:
        return Nothing()


@dataclass(frozen=True)
class MonadPlusMaybe(AlternativeMaybe, MonadMaybe, MonadPlus[Maybe]): ...



from __future__ import annotations

from asyncio import gather
from typing import TypeVar

from categories.type import IO, Lambda

__all__ = (
    'pureIO',
    'bindIO',
    'seqIO',
    'failIO',
    'plusIO',
    'sequenceIO',
)


a = TypeVar('a')

b = TypeVar('b')


async def pureIO(x : a, /) -> a:
    return x


async def bindIO(m : IO[a], k : Lambda[a, IO[b]], /) -> b:
    match await m:
        case x:
            return await k(x)


async def seqIO(m : IO[a], k : IO[b], /) -> b:
    match await m:
        case _:
            return await k


async def failIO(x : str, /) -> a:
    raise Exception(x)


async def plusIO(m : IO[a], m_ : IO[a], /) -> a:
    try:
        return await m
    except BaseException:
        return await m_


async def sequenceIO(xs : list[IO[a]], /) -> list[a]:
    return await gather(*xs)



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.type import typeclass

__all__ = (
    'Show',
)


a = TypeVar('a')


@dataclass(frozen=True)
class Show(typeclass[a]):
    def show(self, x : a, /) -> str: ...



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.control.monad import Monad
from categories.type import hkt, typeclass

__all__ = (
    'MonadTrans',
)


a = TypeVar('a')

m = TypeVar('m')

t = TypeVar('t')


@dataclass(frozen=True)
class MonadTrans(typeclass[t]):
    def lift(self, inst : Monad[m], m : hkt[m, a], /) -> hkt[t, m, a]: ...



from __future__ import annotations

from typing import TypeVar

from categories.data.functor import Functor
from categories.type import Void, hkt

__all__ = (
    'Void',
    'absurd',
    'vacuous',
)


a = TypeVar('a')

f = TypeVar('f')


def absurd(_ : Void, /) -> a: ...


def vacuous(inst : Functor[f], _ : hkt[f, Void], /) -> hkt[f, a]:
    return inst.fmap(absurd, _)


from __future__ import annotations

from typing import TypeVar

from categories.type import Delay, Expr, Map, Void

__all__ = (
    'delay',
    'force',
    'let',
    'rec',
    'seq',
    'undefined',
)


a = TypeVar('a')

b = TypeVar('b')


def delay(x : a, /) -> Delay[a]: return lambda: x


def force(x : Delay[a], /) -> a: return x()


def let(**_ : a) -> Map[str, a]: return Map(_)


def rec(*fs : Expr[..., a]) -> tuple[Expr[..., a], ...]:
    return (x := (*map(lambda f, /: lambda *_: f(*x, *_), fs),))


def seq(_ : a, x : b, /) -> b: return x


def undefined() -> Void: assert None, '⊥'



from __future__ import annotations

from typing import TypeVar

from categories.control.alternative import Alternative
from categories.control.applicative import Applicative
from categories.control.monad import Monad
from categories.control.monadplus import MonadPlus
from categories.data.functor import Functor
from categories.type import Lambda

__all__ = (
    'Parser',
    'FunctorParser',
    'ApplicativeParser',
    'AlternativeParser',
    'MonadParser',
    'MonadPlusParser',
)


a = TypeVar('a')

b = TypeVar('b')


Parser = Lambda[str, list[tuple[a, str]]]


@dataclass(frozen=True)
class FunctorParser(Functor[Parser]):
    def map(self, f : Lambda[a, b], p : Parser[a], /) -> Parser[b]:
        return lambda s, /: [(f(x), s) for (x, s) in p(s)]


@dataclass(frozen=True)
class ApplicativeParser(FunctorParser, Applicative[Parser]):
    def pure(self, x : a, /) -> Parser[a]:
        return lambda s, /: [(x, s)]

    def apply(self, p : Parser[Lambda[a, b]], q : Parser[a], /) -> Parser[b]:
        return lambda s, /: [(f(x), s) for (f, s) in p(s) for (x, s) in q(s)]


@dataclass(frozen=True)
class AlternativeParser(ApplicativeParser, Alternative[Parser]):
    def empty(self, /) -> Parser[a]:
        return lambda _, /: []

    def alt(self, p : Parser[a], q : Parser[a], /) -> Parser[a]:
        return lambda s, /: p(s) or q(s)

    def some(self, p : Parser[a], /) -> Parser[list[a]]:
        return lambda s, /: [([x, *xs], s) for (x, s) in p(s) for (xs, s) in self.many(p)(s)]

    def many(self, p : Parser[a], /) -> Parser[list[a]]:
        return self.alt(self.some(p), self.pure([]))


@dataclass(frozen=True)
class MonadParser(ApplicativeParser, Monad[Parser]):
    def bind(self, p : Parser[a], k : Lambda[a, Parser[b]], /) -> Parser[b]:
        return lambda s, /: [(y, s) for (x, s) in p(s) for (y, s) in k(x)(s)]


@dataclass(frozen=True)
class MonadPlusParser(AlternativeParser, MonadParser, MonadPlus[Parser]): ...



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.control.applicative import Applicative
from categories.control.monad import Monad
from categories.data.functor import Functor
from categories.type import Lambda, forall

__all__ = (
    'Node',
    'FunctorNode',
    'ApplicativeNode',
    'MonadNode',
)


a = TypeVar('a')

b = TypeVar('b')


@dataclass(frozen=True)
class Node(forall[a]):
    x  : a
    xs : list[Node[a]]


@dataclass(frozen=True)
class FunctorNode(Functor[Node]):
    def map(self, f : Lambda[a, b], n : Node[a], /) -> Node[b]:
        match n:
            case Node(x, xs):
                return Node(f(x), [self.map(f, x) for x in xs])

    def const(self, x : a, _ : Node[b], /) -> Node[a]:
        match _:
            case Node(_, xs):
                return Node(x, [self.const(x, _) for _ in xs])


@dataclass(frozen=True)
class ApplicativeNode(FunctorNode, Applicative[Node]):
    def pure(self, x : a, /) -> Node[a]:
        return Node(x, [])

    def apply(self, n : Node[Lambda[a, b]], n_ : Node[a], /) -> Node[b]:
        match n, n_:
            case Node(f, fs), Node(x, xs):
                return Node(f(x), [self.map(f, x) for x in xs] + [self.apply(f, n_) for f in fs])

    def seq(self, _ : Node[a], n : Node[b], /) -> Node[b]:
        match _, n:
            case Node(_, xs), Node(y, ys):
                return Node(y, ys + [self.seq(x, n) for x in xs])


@dataclass(frozen=True)
class MonadNode(ApplicativeNode, Monad[Node]):
    def bind(self, n : Node[a], k : Lambda[a, Node[b]], /) -> Node[b]:
        match n:
            case Node(x, xs):
                match k(x):
                    case Node(y, ys):
                        return Node(y, ys + [self.bind(x, k) for x in xs])



from __future__ import annotations

from dataclasses import dataclass
from functools import partial
from typing import TypeVar

from categories.control.monad import Monad
from categories.data.dual import Dual, MonoidDual
from categories.data.endo import Endo, MonoidEndo
from categories.data.maybe import Just, Maybe, Nothing
from categories.data.monoid import Monoid
from categories.type import Expr, Lambda, hkt, typeclass

__all__ = (
    'Fold',
    'foldrM',
    'foldlM',
)


a = TypeVar('a')

b = TypeVar('b')

m = TypeVar('m')

t = TypeVar('t')


@dataclass(frozen=True)
class Fold(typeclass[t]):
    def fold(self, inst : Monoid[m], xs : hkt[t, m], /) -> m:
        return self.foldMap(inst, lambda x, /: x, xs)

    def foldMap(self, inst : Monoid[m], f : Lambda[a, m], xs : hkt[t, a], /) -> m:
        def g(x : a, y : m, /) -> m:
            return inst.append(f(x), y)
        return self.foldr(g, inst.empty(), xs)

    def foldMap_(self, inst : Monoid[m], f : Lambda[a, m], xs : hkt[t, a], /) -> m:
        def g(x : m, y : a, /) -> m:
            return inst.append(x, f(y))
        return self.foldl_(g, inst.empty(), xs)

    def foldr(self, f : Expr[[a, b], b], z : b, xs : hkt[t, a], /) -> b:
        def g(x : a, /) -> Endo[b]:
            return Endo(partial(f, x))
        return self.foldMap(MonoidEndo(), g, xs).endo(z)

    def foldr_(self, f : Expr[[a, b], b], z : b, xs : hkt[t, a], /) -> b:
        def g(k : Lambda[b, b], x : a, /) -> Lambda[b, b]:
            return lambda y, /: k(f(x, y))
        return self.foldl(g, lambda x, /: x, xs)(z)

    def foldl(self, f : Expr[[b, a], b], z : b, xs : hkt[t, a], /) -> b:
        def g(x : a, /) -> Dual[Endo[b]]:
            return Dual(Endo(lambda y, /: f(y, x)))
        return self.foldMap(MonoidDual(MonoidEndo()), g, xs).dual.endo(z)

    def foldl_(self, f : Expr[[b, a], b], z : b, xs : hkt[t, a], /) -> b:
        def g(x : a, k : Lambda[b, b], /) -> Lambda[b, b]:
            return lambda y, /: k(f(y, x))
        return self.foldr(g, lambda x, /: x, xs)(z)

    def foldr1(self, f : Expr[[a, a], a], xs : hkt[t, a], /) -> a:
        def g(x : a, m : Maybe[a], /) -> Maybe[a]:
            match m:
                case Nothing():
                    return Just(x)
                case Just(y):
                    return Just(f(x, y))

        match self.foldr(g, Nothing(), xs):
            case Nothing():
                assert None
            case Just(x):
                return x

    def foldl1(self, f : Expr[[a, a], a], xs : hkt[t, a], /) -> a:
        def g(m : Maybe[a], y : a, /) -> Maybe[a]:
            match m:
                case Nothing():
                    return Just(y)
                case Just(x):
                    return Just(f(x, y))

        match self.foldl(g, Nothing(), xs):
            case Nothing():
                assert None
            case Just(x):
                return x


def foldrM(fold : Fold[t], monad : Monad[m],
           f : Expr[[a, b], hkt[m, b]], z : b, xs : hkt[t, a], /) -> hkt[m, b]:
    def g(k : Lambda[b, hkt[m, b]], x : a, /) -> Lambda[b, hkt[m, b]]:
        return lambda y, /: monad.bind(f(x, y), k)
    return fold.foldl(g, monad.pure, xs)(z)


def foldlM(fold : Fold[t], monad : Monad[m],
           f : Expr[[b, a], hkt[m, b]], z : b, xs : hkt[t, a], /) -> hkt[m, b]:
    def g(x : a, k : Lambda[b, hkt[m, b]], /) -> Lambda[b, hkt[m, b]]:
        return lambda y, /: monad.bind(f(y, x), k)
    return fold.foldr(g, monad.pure, xs)(z)



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.control.category import Category, CategoryLambda, arrow
from categories.data.tuple import swap
from categories.type import Lambda, hkt, typeclass

__all__ = (
    'Arrow',
    'ArrowLambda',
)


a = TypeVar('a')

b = TypeVar('b')

b_ = TypeVar('b_')

c = TypeVar('c')

c_ = TypeVar('c_')

d = TypeVar('d')


@dataclass(frozen=True)
class Arrow(Category[a], typeclass[a]):
    def arrow(self, f : Lambda[b, c], /) -> hkt[a, b, c]: ...

    def first(self, f : hkt[a, b, c], /) -> hkt[a, tuple[b, d], tuple[c, d]]:
        return self.product(f, self.id())

    def second(self, g : hkt[a, b, c], /) -> hkt[a, tuple[d, b], tuple[d, c]]:
        return self.product(self.id(), g)

    def product(self, f : hkt[a, b, c], g : hkt[a, b_, c_], /) -> hkt[a, tuple[b, b_], tuple[c, c_]]:
        return arrow(self, self.first(f), arrow(self, self.arrow(swap), arrow(self, self.second(g), self.arrow(swap))))

    def fanout(self, f : hkt[a, b, c], g : hkt[a, b, c_], /) -> hkt[a, b, tuple[c, c_]]:
        return arrow(self, self.arrow(lambda x, /: (x, x)), self.product(f, g))


@dataclass(frozen=True)
class ArrowLambda(CategoryLambda, Arrow[Lambda]):
    def arrow(self, f : Lambda[b, c], /) -> Lambda[b, c]:
        return f

    def first(self, f : Lambda[b, c], /) -> Lambda[tuple[b, d], tuple[c, d]]:
        return self.product(f, self.id())

    def second(self, g : Lambda[b, c], /) -> Lambda[tuple[d, b], tuple[d, c]]:
        return self.product(self.id(), g)

    def product(self, f : Lambda[b, c], g : Lambda[b_, c_], /) -> Lambda[tuple[b, b_], tuple[c, c_]]:
        def h(z : tuple[b, b_], /) -> tuple[c, c_]:
            match z:
                case (x, y):
                    return (f(x), g(y))
        return h

    def fanout(self, f : Lambda[b, c], g : Lambda[b, c_], /) -> Lambda[b, tuple[c, c_]]:
        return lambda x, /: (f(x), g(x))



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.control.applicative import Applicative, ApplicativeIO, ApplicativeLambda, ApplicativeList
from categories.type import IO, Lambda, _, hkt, typeclass

__all__ = (
    'Monad',
    'MonadIO',
    'MonadLambda',
    'MonadList',
)


a = TypeVar('a')

b = TypeVar('b')

m = TypeVar('m')

r = TypeVar('r')


@dataclass(frozen=True)
class Monad(Applicative[m], typeclass[m]):
    def bind(self, m : hkt[m, a], k : Lambda[a, hkt[m, b]], /) -> hkt[m, b]:
        return self.join(self.map(k, m))

    def join(self, m : hkt[m, hkt[m, a]], /) -> hkt[m, a]:
        return self.bind(m, lambda x, /: x)

    def seq(self, m : hkt[m, a], k : hkt[m, b], /) -> hkt[m, b]:
        return self.bind(m, lambda _, /: k)


@dataclass(frozen=True)
class MonadIO(ApplicativeIO, Monad[IO]):
    async def bind(self, m : IO[a], k : Lambda[a, IO[b]], /) -> b:
        match await m:
            case x:
                return await k(x)

    async def join(self, m : IO[IO[a]], /) -> a:
        match await m:
            case x:
                return await x


@dataclass(frozen=True)
class MonadLambda(ApplicativeLambda[r], Monad[Lambda[r, _]]):
    def bind(self, f : Lambda[r, a], k : Lambda[a, Lambda[r, b]], /) -> Lambda[r, b]:
        return lambda r, /: k(f(r))(r)


@dataclass(frozen=True)
class MonadList(ApplicativeList, Monad[list]):
    def bind(self, xs : list[a], f : Lambda[a, list[b]], /) -> list[b]:
        return [y for x in xs for y in f(x)]

    def join(self, xss : list[list[a]], /) -> list[a]:
        return [x for xs in xss for x in xs]



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.control.applicative import Applicative
from categories.control.monad import Monad
from categories.data.functor import Functor
from categories.type import Lambda, forall

__all__ = (
    'Id',
    'FunctorId',
    'ApplicativeId',
    'MonadId',
)


a = TypeVar('a')

b = TypeVar('b')


@dataclass(frozen=True)
class Id(forall[a]):
    identity : a


@dataclass(frozen=True)
class FunctorId(Functor[Id]):
    def map(self, f : Lambda[a, b], x : Id[a], /) -> Id[b]:
        match x:
            case Id(x):
                return Id(f(x))


@dataclass(frozen=True)
class ApplicativeId(FunctorId, Applicative[Id]):
    def pure(self, x : a, /) -> Id[a]:
        return Id(x)

    def apply(self, f : Id[Lambda[a, b]], x : Id[a], /) -> Id[b]:
        match f, x:
            case Id(f), Id(x):
                return Id(f(x))


@dataclass(frozen=True)
class MonadId(ApplicativeId, Monad[Id]):
    def bind(self, m : Id[a], k : Lambda[a, Id[b]], /) -> Id[b]:
        match m:
            case Id(x):
                return k(x)



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.control.applicative import Applicative, ApplicativeIO, ApplicativeList
from categories.type import IO, hkt, typeclass

__all__ = (
    'Alternative',
    'AlternativeIO',
    'AlternativeList',
)


a = TypeVar('a')

b = TypeVar('b')

f = TypeVar('f')


@dataclass(frozen=True)
class Alternative(Applicative[f], typeclass[f]):
    def empty(self, /) -> hkt[f, a]: ...

    def alt(self, x : hkt[f, a], y : hkt[f, a], /) -> hkt[f, a]: ...

    def some(self, v : hkt[f, a], /) -> hkt[f, list[a]]:
        return self.apply(self.map(lambda x, /: lambda xs, /: [x, *xs], v), self.many(v))

    def many(self, v : hkt[f, a], /) -> hkt[f, list[a]]:
        return self.alt(self.some(v), self.pure([]))


@dataclass(frozen=True)
class AlternativeIO(ApplicativeIO, Alternative[IO]):
    async def empty(self, /) -> a:
        raise Exception

    async def alt(self, m : IO[a], m_ : IO[a], /) -> a:
        try:
            return await m
        except BaseException:
            return await m_


@dataclass(frozen=True)
class AlternativeList(ApplicativeList, Alternative[list]):
    def empty(self, /) -> list[a]:
        return []

    def alt(self, xs : list[a], ys : list[a], /) -> list[a]:
        return xs + ys



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.control.applicative import Applicative
from categories.data.fold import Fold
from categories.data.functor import Functor
from categories.type import Lambda, hkt, typeclass

__all__ = (
    'Traverse',
)


a = TypeVar('a')

b = TypeVar('b')

f = TypeVar('f')

t = TypeVar('t')


@dataclass(frozen=True)
class Traverse(Functor[t], Fold[t], typeclass[t]):
    def traverse(self, inst : Applicative[f],
                 f : Lambda[a, hkt[f, b]], xs : hkt[t, a], /) -> hkt[f, hkt[t, b]]:
        return self.sequence(inst, self.map(f, xs))

    def sequence(self, inst : Applicative[f],
                  xs : hkt[t, hkt[f, a]], /) -> hkt[f, hkt[t, a]]:
        return self.traverse(inst, lambda x, /: x, xs)



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.control.alternative import Alternative
from categories.control.applicative import Applicative
from categories.control.liftio import LiftIO
from categories.control.monad import Monad
from categories.control.monadplus import MonadPlus
from categories.control.monadtrans import MonadTrans
from categories.data.functor import Functor
from categories.type import IO, Lambda, _, forall, hkt

__all__ = (
    'StateT',
    'FunctorStateT',
    'ApplicativeStateT',
    'AlternativeStateT',
    'MonadStateT',
    'MonadPlusStateT',
    'MonadTransStateT',
    'LiftIOStateT',
)


a = TypeVar('a')

b = TypeVar('b')

m = TypeVar('m')

s = TypeVar('s')


@dataclass(frozen=True)
class StateT(forall[s, m, a]):
    run : Lambda[s, hkt[m, tuple[a, s]]]


@dataclass(frozen=True)
class FunctorStateT(Functor[StateT[s, m, _]]):
    inst : Functor[m]

    def map(self, f : Lambda[a, b], x : StateT[s, m, a], /) -> StateT[s, m, b]:
        match self:
            case FunctorStateT(inst):
                def g(_ : tuple[a, s], /) -> tuple[b, s]:
                    match _:
                        case (x, s_):
                            return (f(x), s_)
                return StateT(lambda s, /: inst.map(g, x.run(s)))


@dataclass(frozen=True)
class ApplicativeStateT(FunctorStateT[s, m], Applicative[StateT[s, m, _]]):
    inst : Monad[m]

    def pure(self, x : a, /) -> StateT[s, m, a]:
        match self:
            case ApplicativeStateT(inst):
                return StateT(lambda s, /: inst.pure((x, s)))

    def apply(self, f : StateT[s, m, Lambda[a, b]], x : StateT[s, m, a], /) -> StateT[s, m, b]:
        match self:
            case ApplicativeStateT(inst):
                def k(_ : tuple[Lambda[a, b], s], /) -> hkt[m, tuple[b, s]]:
                    match _:
                        case (f, s_):
                            def k_(_ : tuple[a, s], /) -> hkt[m, tuple[b, s]]:
                                match _:
                                    case (x, s__):
                                        return inst.pure((f(x), s__))
                            return inst.bind(x.run(s_), k_)
                return StateT(lambda s, /: inst.bind(f.run(s), k))


@dataclass(frozen=True)
class AlternativeStateT(ApplicativeStateT[s, m], Alternative[StateT[s, m, _]]):
    inst : Alternative[m]

    def empty(self, /) -> StateT[s, m, a]:
        match self:
            case AlternativeStateT(inst):
                return StateT(lambda _, /: inst.empty())

    def alt(self, x : StateT[s, m, a], y : StateT[s, m, a], /) -> StateT[s, m, a]:
        match self:
            case AlternativeStateT(inst):
                return StateT(lambda s, /: inst.alt(x.run(s), y.run(s)))


@dataclass(frozen=True)
class MonadStateT(ApplicativeStateT[s, m], Monad[StateT[s, m, _]]):
    inst : Monad[m]

    def bind(self, m : StateT[s, m, a], k : Lambda[a, StateT[s, m, b]], /) -> StateT[s, m, b]:
        match self:
            case MonadStateT(inst):
                def k_(_ : tuple[a, s], /) -> hkt[m, tuple[b, s]]:
                    match _:
                        case (x, s_):
                            return k(x).run(s_)
                return StateT(lambda s, /: inst.bind(m.run(s), k_))


@dataclass(frozen=True)
class MonadPlusStateT(AlternativeStateT[s, m], MonadStateT[s, m], MonadPlus[StateT[s, m, _]]):
    inst : MonadPlus[m]


@dataclass(frozen=True)
class MonadTransStateT(MonadTrans[StateT[s, _, _]]):
    def lift(self, inst : Monad[m], m : hkt[m, a], /) -> StateT[s, m, a]:
        return StateT(lambda s, /: inst.bind(m, lambda x, /: inst.pure((x, s))))


@dataclass(frozen=True)
class LiftIOStateT(MonadStateT[s, m], LiftIO[StateT[s, m, _]]):
    inst : LiftIO[m]

    def liftIO(self, m : IO[a], /) -> StateT[s, m, a]:
        match self:
            case LiftIOStateT(inst):
                return StateT(lambda s, /: inst.bind(inst.liftIO(m), lambda x, /: inst.pure((x, s))))



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.type import IO, Lambda, _, hkt, typeclass

__all__ = (
    'Functor',
    'FunctorIO',
    'FunctorLambda',
    'FunctorList',
)


a = TypeVar('a')

b = TypeVar('b')

f = TypeVar('f')

r = TypeVar('r')


@dataclass(frozen=True)
class Functor(typeclass[f]):
    def map(self, f : Lambda[a, b], x : hkt[f, a], /) -> hkt[f, b]: ...

    def const(self, x : a, _ : hkt[f, b], /) -> hkt[f, a]:
        return self.map(lambda _, /: x, _)


@dataclass(frozen=True)
class FunctorIO(Functor[IO]):
    async def map(self, f : Lambda[a, b], m : IO[a], /) -> b:
        match await m:
            case x:
                return f(x)

    async def const(self, x : a, _ : IO[b], /) -> a:
        match await _:
            case _:
                return x


@dataclass(frozen=True)
class FunctorLambda(Functor[Lambda[r, _]]):
    def map(self, f : Lambda[a, b], g : Lambda[r, a], /) -> Lambda[r, b]:
        return lambda x, /: f(g(x))


@dataclass(frozen=True)
class FunctorList(Functor[list]):
    def map(self, f : Lambda[a, b], xs : list[a], /) -> list[b]:
        return [f(x) for x in xs]



from __future__ import annotations

from typing import TypeVar

from categories.type import Expr, Lambda

__all__ = (
    'cons',
    'head',
    'tail',
    'filter',
    'foldl',
    'foldr',
    'map',
    'scanl',
    'scanr',
    'unfoldr',
)


a = TypeVar('a')

b = TypeVar('b')


def cons(x : a, xs : list[a], /) -> list[a]:
    return [x, *xs]


def head(xs : list[a], /) -> a:
    match xs:
        case []:
            assert None
        case [x, *_]:
            return x


def tail(xs : list[a], /) -> list[a]:
    match xs:
        case []:
            return []
        case [_, *xs]:
            return xs


def filter(f : Lambda[a, bool], xs : list[a], /) -> list[a]:
    return [x for x in xs if f(x)]


def foldl(f : Expr[[b, a], b], z : b, xs : list[a], /) -> b:
    match xs:
        case []:
            return z
        case [x, *xs]:
            return foldl(f, f(z, x), xs)


def foldr(f : Expr[[a, b], b], z : b, xs : list[a], /) -> b:
    match xs:
        case []:
            return z
        case [x, *xs]:
            return f(x, foldr(f, z, xs))


def map(f : Lambda[a, b], xs : list[a], /) -> list[b]:
    return [f(x) for x in xs]


def scanl(f : Expr[[b, a], b], z : b, xs : list[a], /) -> list[b]:
    match xs:
        case []:
            return [z]
        case [x, *xs]:
            return [z, *scanl(f, f(z, x), xs)]


def scanr(f : Expr[[a, b], b], z : b, xs : list[a], /) -> list[b]:
    match xs:
        case []:
            return [z]
        case [x, *xs]:
            match scanr(f, z, xs):
                case [y, *_] as ys:
                    return [f(x, y), *ys]


def unfoldr(f : Lambda[b, None | tuple[a, b]], z : b, /) -> list[a]:
    match f(z):
        case None:
            return []
        case (x, y):
            return [x, *unfoldr(f, y)]



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.control.applicative import Applicative
from categories.control.monad import Monad
from categories.data.functor import Functor
from categories.type import Lambda, forall

__all__ = (
    'Either',
    'Left',
    'Right',
    'FunctorEither',
    'ApplicativeEither',
    'MonadEither',
)


a = TypeVar('a')

b = TypeVar('b')

e = TypeVar('e')


@dataclass(frozen=True)
class Left(forall[a]):
    x : a


@dataclass(frozen=True)
class Right(forall[b]):
    y : b


Either = Left[a] | Right[b]


@dataclass(frozen=True)
class FunctorEither(Functor[Either[e, _]]):
    def map(self, f : Lambda[a, b], e : Either[e, a], /) -> Either[e, b]:
        match e:
            case Left(x):
                return Left(x)
            case Right(y):
                return Right(f(y))


@dataclass(frozen=True)
class ApplicativeEither(FunctorEither, Applicative[Either[e, _]]):
    def pure(self, x : a, /) -> Either[e, a]:
        return Right(x)

    def apply(self, e : Either[e, Lambda[a, b]], e_ : Either[e, a], /) -> Either[e, b]:
        match e:
            case Left(x):
                return Left(x)
            case Right(f):
                return self.map(f, e_)


@dataclass(frozen=True)
class MonadEither(ApplicativeEither, Monad[Either[e, _]]):
    def bind(self, e : Either[e, a], k : Lambda[a, Either[e, b]], /) -> Either[e, b]:
        match e:
            case Left(x):
                return Left(x)
            case Right(y):
                return k(y)



from __future__ import annotations

from typing import TypeVar

from categories.type import Expr, Lambda

__all__ = (
    'Curry',
    'fst',
    'snd',
    'curry',
    'uncurry',
    'swap',
)


a = TypeVar('a')

b = TypeVar('b')

c = TypeVar('c')


Curry = Lambda[a, Lambda[b, c]]


def fst(z : tuple[a, b], /) -> a:
    match z:
        case (x, _):
            return x


def snd(z : tuple[a, b], /) -> b:
    match z:
        case (_, y):
            return y


def curry(f : Expr[[a, b], c], /) -> Curry[a, b, c]:
    return lambda x, /: lambda y, /: f(x, y)


def uncurry(f : Curry[a, b, c], /) -> Expr[[a, b], c]:
    return lambda x, y, /: f(x)(y)


def swap(z : tuple[a, b], /) -> tuple[b, a]:
    match z:
        case (x, y):
            return (y, x)



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.control.monad import Monad, MonadIO, MonadList
from categories.type import IO, hkt, typeclass

__all__ = (
    'MonadFail',
    'MonadFailIO',
    'MonadFailList',
)


a = TypeVar('a')

m = TypeVar('m')


@dataclass(frozen=True)
class MonadFail(Monad[m], typeclass[m]):
    def fail(self, x : str, /) -> hkt[m, a]: ...


@dataclass(frozen=True)
class MonadFailIO(MonadIO, MonadFail[IO]):
    async def fail(self, x : str, /) -> a:
        raise Exception(x)


@dataclass(frozen=True)
class MonadFailList(MonadList, MonadFail[list]):
    def fail(self, _ : str, /) -> list[a]:
        return []



from __future__ import annotations

from dataclasses import dataclass
from functools import reduce
from itertools import repeat
from typing import TypeVar

from categories.data.function import o
from categories.data.monoid import Monoid
from categories.data.semigroup import Semigroup
from categories.type import Lambda, forall

__all__ = (
    'Endo',
    'SemigroupEndo',
    'MonoidEndo',
)


a = TypeVar('a')


@dataclass(frozen=True)
class Endo(forall[a]):
    endo : Lambda[a, a]


@dataclass(frozen=True)
class SemigroupEndo(Semigroup[Endo[a]]):
    def append(self, x : Endo[a], y : Endo[a], /) -> Endo[a]:
        match x, y:
            case Endo(f), Endo(g):
                return Endo(o(f, g))

    def times(self, n : int, x : Endo[a], /) -> Endo[a]:
        match x:
            case Endo(f):
                return Endo(reduce(o, repeat(f, n)))


@dataclass(frozen=True)
class MonoidEndo(SemigroupEndo[a], Monoid[Endo[a]]):
    def empty(self, /) -> Endo[a]:
        return Endo(lambda x, /: x)



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.data.eq import Eq
from categories.type import typeclass

__all__ = (
    'Ord',
    'Ordering',
    'LT',
    'EQ',
    'GT',
)


a = TypeVar('a')


@dataclass(frozen=True)
class LT: ...


@dataclass(frozen=True)
class EQ: ...


@dataclass(frozen=True)
class GT: ...


Ordering = LT | EQ | GT


@dataclass(frozen=True)
class Ord(Eq[a], typeclass[a]):
    def cmp(self, x : a, y : a, /) -> Ordering:
        return EQ() if self.eq(x, y) \
          else LT() if self.le(x, y) \
          else GT()

    def le(self, x : a, y : a, /) -> bool:
        match self.cmp(x, y):
            case GT():
                return False
            case _:
                return True

    def ge(self, x : a, y : a, /) -> bool:
        return self.le(y, x)

    def gt(self, x : a, y : a, /) -> bool:
        return not self.le(x, y)

    def lt(self, x : a, y : a, /) -> bool:
        return not self.le(y, x)

    def max(self, x : a, y : a, /) -> a:
        return y if self.le(x, y) else x

    def min(self, x : a, y : a, /) -> a:
        return x if self.le(x, y) else y



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.control.alternative import Alternative
from categories.control.applicative import Applicative
from categories.control.liftio import LiftIO
from categories.control.monad import Monad
from categories.control.monadplus import MonadPlus
from categories.control.monadtrans import MonadTrans
from categories.data.functor import Functor
from categories.data.monoid import Monoid
from categories.type import IO, Lambda, _, forall, hkt

__all__ = (
    'WriterT',
    'FunctorWriterT',
    'ApplicativeWriterT',
    'AlternativeWriterT',
    'MonadWriterT',
    'MonadPlusWriterT',
    'MonadTransWriterT',
    'LiftIOWriterT',
)


a = TypeVar('a')

b = TypeVar('b')

m = TypeVar('m')

w = TypeVar('w')


@dataclass(frozen=True)
class WriterT(forall[w, m, a]):
    run : hkt[m, tuple[a, w]]


@dataclass(frozen=True)
class FunctorWriterT(Functor[WriterT[w, m, _]]):
    monoid : Monoid[w]
    inst : Functor[m]

    def map(self, f : Lambda[a, b], x : WriterT[w, m, a], /) -> WriterT[w, m, b]:
        match self:
            case FunctorWriterT(monoid, inst):
                def g(_ : tuple[a, w], /) -> tuple[b, w]:
                    match _:
                        case (x, w):
                            return (f(x), w)
                return WriterT(inst.map(g, x.run))


@dataclass(frozen=True)
class ApplicativeWriterT(FunctorWriterT[w, m], Applicative[WriterT[w, m, _]]):
    monoid : Monoid[w]
    inst : Applicative[m]

    def pure(self, x : a, /) -> WriterT[w, m, a]:
        match self:
            case ApplicativeWriterT(monoid, inst):
                return WriterT(inst.pure((x, monoid.empty())))

    def apply(self, f : WriterT[w, m, Lambda[a, b]], x : WriterT[w, m, a], /) -> WriterT[w, m, b]:
        match self:
            case ApplicativeWriterT(monoid, inst):
                def g(_ : tuple[Lambda[a, b], w], /) -> tuple[b, w]:
                    match _:
                        case (f, w):
                            def h(_ : tuple[a, w], /) -> tuple[b, w]:
                                match _:
                                    case (x, w_):
                                        return (f(x), monoid.append(w, w_))
                            return h
                return WriterT(inst.apply(inst.map(g, f.run), x.run))


@dataclass(frozen=True)
class AlternativeWriterT(ApplicativeWriterT[w, m], Alternative[WriterT[w, m, _]]):
    monoid : Monoid[w]
    inst : Alternative[m]

    def empty(self, /) -> WriterT[w, m, a]:
        match self:
            case AlternativeWriterT(monoid, inst):
                return WriterT(inst.empty())

    def alt(self, x : WriterT[w, m, a], y : WriterT[w, m, a], /) -> WriterT[w, m, a]:
        match self:
            case AlternativeWriterT(monoid, inst):
                return WriterT(inst.alt(x.run, y.run))


@dataclass(frozen=True)
class MonadWriterT(ApplicativeWriterT[w, m], Monad[WriterT[w, m, _]]):
    monoid : Monoid[w]
    inst : Monad[m]

    def bind(self, m : WriterT[w, m, a], k : Lambda[a, WriterT[w, m, b]], /) -> WriterT[w, m, b]:
        match self:
            case MonadWriterT(monoid, inst):
                def k_(_ : tuple[a, w], /) -> hkt[m, tuple[b, w]]:
                    match _:
                        case (x, w):
                            def k__(_ : tuple[b, w], /) -> hkt[m, tuple[b, w]]:
                                match _:
                                    case (y, w_):
                                        return inst.pure((y, monoid.append(w, w_)))
                            return inst.bind(k(x).run, k__)
                return WriterT(inst.bind(m.run, k_))


@dataclass(frozen=True)
class MonadPlusWriterT(AlternativeWriterT[w, m], MonadWriterT[w, m], MonadPlus[WriterT[w, m, _]]):
    monoid : Monoid[w]
    inst : MonadPlus[m]


@dataclass(frozen=True)
class MonadTransWriterT(MonadTrans[WriterT[w, _, _]]):
    monoid : Monoid[w]

    def lift(self, inst : Monad[m], m : hkt[m, a], /) -> WriterT[w, m, a]:
        match self:
            case MonadTransWriterT(monoid):
                return WriterT(inst.bind(m, lambda x, /: inst.pure((x, monoid.empty()))))


@dataclass(frozen=True)
class LiftIOWriterT(MonadWriterT[w, m], LiftIO[WriterT[w, m, _]]):
    monoid : Monoid[w]
    inst : LiftIO[m]

    def liftIO(self, m : IO[a], /) -> WriterT[w, m, a]:
        match self:
            case LiftIOWriterT(monoid, inst):
                return WriterT(inst.bind(inst.liftIO(m), lambda x, /: inst.pure((x, monoid.empty()))))



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.data.functor import Functor, FunctorIO, FunctorLambda, FunctorList
from categories.type import IO, Lambda, _, hkt, typeclass

__all__ = (
    'Applicative',
    'ApplicativeIO',
    'ApplicativeLambda',
    'ApplicativeList',
)


a = TypeVar('a')

b = TypeVar('b')

f = TypeVar('f')

r = TypeVar('r')


@dataclass(frozen=True)
class Applicative(Functor[f], typeclass[f]):
    def pure(self, x : a, /) -> hkt[f, a]: ...

    def apply(self, f : hkt[f, Lambda[a, b]], x : hkt[f, a], /) -> hkt[f, b]: ...

    def seq(self, _ : hkt[f, a], x : hkt[f, b], /) -> hkt[f, b]:
        return self.apply(self.const(lambda x, /: x, _), x)


@dataclass(frozen=True)
class ApplicativeIO(FunctorIO, Applicative[IO]):
    async def pure(self, x : a, /) -> a:
        return x

    async def apply(self, m : IO[Lambda[a, b]], m_ : IO[a], /) -> b:
        match await m, await m_:
            case f, x:
                return f(x)

    async def seq(self, m : IO[a], k : IO[b], /) -> b:
        match await m:
            case _:
                return await k


@dataclass(frozen=True)
class ApplicativeLambda(FunctorLambda[r], Applicative[Lambda[r, _]]):
    def pure(self, x : a, /) -> Lambda[r, a]:
        return lambda _, /: x

    def apply(self, f : Lambda[r, Lambda[a, b]], g : Lambda[r, a], /) -> Lambda[r, b]:
        return lambda x, /: f(x)(g(x))


@dataclass(frozen=True)
class ApplicativeList(FunctorList, Applicative[list]):
    def pure(self, x : a, /) -> list[a]:
        return [x]

    def apply(self, fs : list[Lambda[a, b]], xs : list[a], /) -> list[b]:
        return [f(x) for f in fs for x in xs]



from __future__ import annotations

from dataclasses import dataclass
from functools import reduce
from itertools import repeat
from typing import TypeVar

from categories.type import typeclass

__all__ = (
    'Semigroup',
)


a = TypeVar('a')


@dataclass(frozen=True)
class Semigroup(typeclass[a]):
    def append(self, x : a, y : a, /) -> a:
        return self.concat([x, y])

    def concat(self, xs : list[a], /) -> a:
        return reduce(self.append, xs)

    def times(self, n : int, x : a, /) -> a:
        return reduce(self.append, repeat(x, n))



from __future__ import annotations

from builtins import frozenset as Set
from collections.abc import Callable as Expr
from collections.abc import Coroutine as Action
from collections.abc import Iterator as Stream
from types import MappingProxyType as Map
from typing import Annotated as hkt
from typing import Any as _
from typing import Generic as forall
from typing import NoReturn as Void
from typing import Protocol as typeclass
from typing import TypeVar

__all__ = (
    'Delay',
    'Action',
    'Expr',
    'IO',
    'Lambda',
    'Map',
    'Set',
    'Stream',
    'Void',
    '_',
    'forall',
    'hkt',
    'typeclass',
)


a = TypeVar('a')

b = TypeVar('b')


Delay = Expr[[], a]

IO = Action[_, _, a]

Lambda = Expr[[a], b]



from __future__ import annotations

from dataclasses import dataclass
from itertools import count
from typing import TypeVar

from categories.type import Stream, typeclass

__all__ = (
    'Bounded',
    'Enum',
    'boundedEnumFrom',
    'boundedEnumFromThen',
)


a = TypeVar('a')


@dataclass(frozen=True)
class Bounded(typeclass[a]):
    def min(self, /) -> a: ...

    def max(self, /) -> a: ...


@dataclass(frozen=True)
class Enum(typeclass[a]):
    def succ(self, x : a, /) -> a:
        return self.toEnum(self.fromEnum(x) + 1)

    def pred(self, x : a, /) -> a:
        return self.toEnum(self.fromEnum(x) - 1)

    def toEnum(self, x : int, /) -> a: ...

    def fromEnum(self, x : a, /) -> int: ...

    def enumFrom(self, x : a, /) -> Stream[a]:
        return map(self.toEnum, count(self.fromEnum(x)))

    def enumFromThen(self, x : a, y : a, /) -> Stream[a]:
        return map(self.toEnum, count(self.fromEnum(x), self.fromEnum(y) - self.fromEnum(x)))

    def enumFromTo(self, x : a, y : a, /) -> Stream[a]:
        return map(self.toEnum, range(self.fromEnum(x), self.fromEnum(y) + 1))

    def enumFromThenTo(self, x : a, y : a, z : a, /) -> Stream[a]:
        return map(self.toEnum, range(self.fromEnum(x), self.fromEnum(z) + 1, self.fromEnum(y) - self.fromEnum(x)))


def boundedEnumFrom(enum : Enum[a], bounded : Bounded[a], x : a, /) -> Stream[a]:
    return map(enum.toEnum, range(enum.fromEnum(x), enum.fromEnum(bounded.max()) + 1))


def boundedEnumFromThen(enum : Enum[a], bounded : Bounded[a], x : a, y : a, /) -> Stream[a]:
    match enum.fromEnum(x), enum.fromEnum(y):
        case x_, y_ if y_ >= x_:
            return map(enum.toEnum, range(x_, enum.fromEnum(bounded.max()) + 1, y_ - x_))
        case x_, y_:
            return map(enum.toEnum, range(x_, enum.fromEnum(bounded.min()) - 1, y_ - x_))



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.type import typeclass

__all__ = (
    'Eq',
)


a = TypeVar('a')


@dataclass(frozen=True)
class Eq(typeclass[a]):
    def eq(self, x : a, y : a, /) -> bool:
        return not self.ne(x, y)

    def ne(self, x : a, y : a, /) -> bool:
        return not self.eq(x, y)



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.data.monoid import Monoid
from categories.data.semigroup import Semigroup
from categories.type import forall

__all__ = (
    'Dual',
    'SemigroupDual',
    'MonoidDual',
)


a = TypeVar('a')


@dataclass(frozen=True)
class Dual(forall[a]):
    dual : a


@dataclass(frozen=True)
class SemigroupDual(Semigroup[Dual[a]]):
    inst : Semigroup[a]

    def append(self, x : Dual[a], y : Dual[a], /) -> Dual[a]:
        match self, x, y:
            case SemigroupDual(inst), Dual(x_), Dual(y_):
                return Dual(inst.append(y_, x_))

    def times(self, n : int, x : Dual[a], /) -> Dual[a]:
        match self, x:
            case SemigroupDual(inst), Dual(x_):
                return Dual(inst.times(n, x_))


@dataclass(frozen=True)
class MonoidDual(SemigroupDual[a], Monoid[Dual[a]]):
    inst : Monoid[a]

    def empty(self, /) -> Dual[a]:
        match self:
            case MonoidDual(inst):
                return Dual(inst.empty())



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

from categories.control.alternative import Alternative
from categories.control.applicative import Applicative
from categories.control.liftio import LiftIO
from categories.control.monad import Monad
from categories.control.monadplus import MonadPlus
from categories.control.monadtrans import MonadTrans
from categories.data.functor import Functor
from categories.data.maybe import Maybe, Nothing, Just
from categories.type import IO, Lambda, _, forall, hkt

__all__ = (
    'MaybeT',
    'FunctorMaybeT',
    'ApplicativeMaybeT',
    'AlternativeMaybeT',
    'MonadMaybeT',
    'MonadPlusMaybeT',
    'MonadTransMaybeT',
    'LiftIOMaybeT',
)


a = TypeVar('a')

b = TypeVar('b')

m = TypeVar('m')


@dataclass(frozen=True)
class MaybeT(forall[m, a]):
    run : hkt[m, Maybe[a]]


@dataclass(frozen=True)
class FunctorMaybeT(Functor[MaybeT[m, _]]):
    inst : Functor[m]

    def map(self, f : Lambda[a, b], x : MaybeT[m, a], /) -> MaybeT[m, b]:
        match self:
            case FunctorMaybeT(inst):
                def g(m : Maybe[a], /) -> Maybe[b]:
                    match m:
                        case Nothing():
                            return Nothing()
                        case Just(x):
                            return Just(f(x))
                return MaybeT(inst.map(g, x.run))


@dataclass(frozen=True)
class ApplicativeMaybeT(FunctorMaybeT[m], Applicative[MaybeT[m, _]]):
    inst : Monad[m]

    def pure(self, x : a, /) -> MaybeT[m, a]:
        match self:
            case ApplicativeMaybeT(inst):
                return MaybeT(inst.pure(Just(x)))

    def apply(self, f : MaybeT[m, Lambda[a, b]], x : MaybeT[m, a], /) -> MaybeT[m, b]:
        match self:
            case ApplicativeMaybeT(inst):
                def k(m : Maybe[Lambda[a, b]], /) -> hkt[m, Maybe[b]]:
                    match m:
                        case Nothing():
                            return inst.pure(Nothing())
                        case Just(f):
                            def k_(m_ : Maybe[a], /) -> hkt[m, Maybe[b]]:
                                match m_:
                                    case Nothing():
                                        return inst.pure(Nothing())
                                    case Just(x):
                                        return inst.pure(Just(f(x)))
                            return inst.bind(x.run, k_)
                return MaybeT(inst.bind(f.run, k))


@dataclass(frozen=True)
class AlternativeMaybeT(ApplicativeMaybeT[m], Alternative[MaybeT[m, _]]):
    inst : Monad[m]

    def empty(self, /) -> MaybeT[m, a]:
        match self:
            case AlternativeMaybeT(inst):
                return MaybeT(inst.pure(Nothing()))

    def alt(self, x : MaybeT[m, a], y : MaybeT[m, a], /) -> MaybeT[m, a]:
        match self:
            case AlternativeMaybeT(inst):
                def k(m : Maybe[a], /) -> hkt[m, Maybe[a]]:
                    match m:
                        case Nothing():
                            return y.run
                        case Just(_):
                            return inst.pure(m)
                return MaybeT(inst.bind(x.run, k))


@dataclass(frozen=True)
class MonadMaybeT(ApplicativeMaybeT[m], Monad[MaybeT[m, _]]):
    inst : Monad[m]

    def bind(self, m : MaybeT[m, a], k : Lambda[a, MaybeT[m, b]], /) -> MaybeT[m, b]:
        match self:
            case MonadMaybeT(inst):
                def k_(m_ : Maybe[a], /) -> hkt[m, Maybe[b]]:
                    match m_:
                        case Nothing():
                            return inst.pure(Nothing())
                        case Just(x):
                            return k(x).run
                return MaybeT(inst.bind(m.run, k_))


@dataclass(frozen=True)
class MonadPlusMaybeT(AlternativeMaybeT[m], MonadMaybeT[m], MonadPlus[MaybeT[m, _]]):
    inst : Monad[m]


@dataclass(frozen=True)
class MonadTransMaybeT(MonadTrans[MaybeT]):
    def lift(self, inst : Monad[m], m : hkt[m, a], /) -> MaybeT[m, a]:
        return MaybeT(inst.map(Just, m))


@dataclass(frozen=True)
class LiftIOMaybeT(MonadMaybeT[m], LiftIO[MaybeT[m, _]]):
    inst : LiftIO[m]

    def liftIO(self, m : IO[a], /) -> MaybeT[m, a]:
        match self:
            case LiftIOMaybeT(inst):
                return MaybeT(inst.map(Just, inst.liftIO(m)))
